@@ -1,5 +1,5 @@
 /*  Copyright 1986-1992 Emmet P. Gray.
- *  Copyright 1994,1996-2002,2007-2009 Alain Knaff.
+ *  Copyright 1994,1996-2002,2007-2009,2021-2022 Alain Knaff.
  *  This file is part of mtools.
  *
  *  Mtools is free software: you can redistribute it and/or modify
@@ -22,16 +22,16 @@
 
 
 #include "sysincludes.h"
-#include "msdos.h"
 #include "mtools.h"
-#include "vfat.h"
 #include "mainloop.h"
 #include "plain_io.h"
 #include "nameclash.h"
 #include "file.h"
 #include "fs.h"
 
-
+#if defined(HAVE_UTIMES) && defined(HAVE_SYS_TIME_H)
+#include <sys/time.h>
+#endif
 /*
  * Preserve the file modification times after the fclose()
  */
@@ -40,14 +40,20 @@ static void set_mtime(const char *target, time_t mtime)
 {
 	if (target && strcmp(target, "-") && mtime != 0L) {
 #ifdef HAVE_UTIMES
-		struct timeval tv[2];	
+		struct timeval tv[2];
 		tv[0].tv_sec = mtime;
 		tv[0].tv_usec = 0;
 		tv[1].tv_sec = mtime;
 		tv[1].tv_usec = 0;
-		utimes((char *)target, tv);
+		utimes(target, tv);
 #else
 #ifdef HAVE_UTIME
+#ifndef HAVE_UTIMBUF
+		struct utimbuf {
+			time_t actime;       /* access time */
+			time_t modtime;      /* modification time */
+		};
+#endif		
 		struct utimbuf utbuf;
 
 		utbuf.actime = mtime;
@@ -74,8 +80,91 @@ typedef struct Arg_t {
 	MainParam_t mp;
 	ClashHandling_t ch;
 	int noClobber;
+
+	const char *unixTarget; /* directory on Unix where to put files,
+				 * needed by mcopy */
 } Arg_t;
 
+static char *buildUnixFilename(Arg_t *arg)
+{
+	const char *target;
+	char *ret;
+	char *tmp;
+
+	target = mpPickTargetName(&arg->mp);
+	ret = malloc(strlen(arg->unixTarget) + 2 + strlen(target));
+	if(!ret)
+		return 0;
+	strcpy(ret, arg->unixTarget);
+	strcat(ret, "/");
+	if(*target) {
+		if(!strcmp(target, ".")) {
+		  target="DOT";
+		} else if(!strcmp(target, "..")) {
+		  target="DOTDOT";
+		}
+		while( (tmp=strchr(target, '/')) ) {
+		  strncat(ret, target, ptrdiff(tmp,target));
+		  strcat(ret, "\\");
+		  target=tmp+1;
+		}
+		strcat(ret, target);
+	}
+	return ret;
+}
+
+/*
+ * Is target a Unix directory
+ * -1 error occured
+ * 0 regular file
+ * 1 directory
+ */
+static int unix_is_dir(const char *name)
+{
+	struct stat buf;
+	if(stat(name, &buf) < 0)
+		return -1;
+	else
+		return 1 && S_ISDIR(buf.st_mode);
+}
+
+static int unix_target_lookup(Arg_t *arg, const char *in)
+{
+	char *ptr;
+	arg->unixTarget = strdup(in);
+	/* try complete filename */
+	if(access(arg->unixTarget, F_OK) == 0) {
+		switch(unix_is_dir(arg->unixTarget)) {
+		case -1:
+			return ERROR_ONE;
+		case 0:
+			break;
+		default:
+			return GOT_ONE;
+		}
+	} else if(errno != ENOENT)
+		return ERROR_ONE;
+
+	ptr = strrchr(arg->unixTarget, '/');
+	if(!ptr) {
+		arg->mp.targetName = arg->unixTarget;
+		arg->unixTarget = strdup(".");
+		return GOT_ONE;
+	} else {
+		*ptr = '\0';
+		arg->mp.targetName = ptr+1;
+		return GOT_ONE;
+	}
+}
+
+int target_lookup(Arg_t *arg, const char *in)
+{
+	if(in[0] && in[1] == ':' )
+		return dos_target_lookup(&arg->mp, in);
+	else
+		return unix_target_lookup(arg, in);
+}
+
 static int _unix_write(MainParam_t *mp, int needfilter, const char *unixFile);
 
 /* Write the Unix file */
@@ -86,7 +175,7 @@ static int unix_write(MainParam_t *mp, int needfilter)
 	if(arg->type)
 		return _unix_write(mp, needfilter, "-");
 	else {
-		char *unixFile = mpBuildUnixFilename(mp);
+		char *unixFile = buildUnixFilename(arg);
 		int ret;
 		if(!unixFile) {
 			printOom();
@@ -107,7 +196,6 @@ static int _unix_write(MainParam_t *mp, int needfilter, const char *unixFile)
 	Stream_t *File=mp->File;
 	Stream_t *Target, *Source;
 	struct MT_STAT stbuf;
-	ssize_t ret;
 	char errmsg[80];
 
 	File->Class->get_data(File, &mtime, 0, 0, 0);
@@ -130,16 +218,14 @@ static int _unix_write(MainParam_t *mp, int needfilter, const char *unixFile)
 				if(!S_ISREG(stbuf.st_mode)) {
 					fprintf(stderr,"\"%s\" is not a regular file\n",
 						unixFile);
-				
+
 					return ERROR_ONE;
 				}
 				sFd = get_fd(File);
 				if(sFd == -1) {
-					fprintf(stderr, "Not ok Unix file ==> good\n");
-				}
-				if((!MT_FSTAT(sFd, &srcStbuf)) &&
-				   stbuf.st_dev == srcStbuf.st_dev &&
-				   stbuf.st_ino == srcStbuf.st_ino) {
+				} else if((!MT_FSTAT(sFd, &srcStbuf)) &&
+					   stbuf.st_dev == srcStbuf.st_dev &&
+					   stbuf.st_ino == srcStbuf.st_ino) {
 					fprintf(stderr, "Attempt to copy file on itself\n");
 					return ERROR_ONE;
 				}
@@ -149,7 +235,7 @@ static int _unix_write(MainParam_t *mp, int needfilter, const char *unixFile)
 					     unixFile)) {
 				return ERROR_ONE;
 			}
-			
+
 		}
 	}
 
@@ -158,7 +244,7 @@ static int _unix_write(MainParam_t *mp, int needfilter, const char *unixFile)
 		mpPrintFilename(stderr,mp);
 		fprintf(stderr,"\n");
 	}
-	
+
 	if(got_signal) {
 		return ERROR_ONE;
 	}
@@ -166,19 +252,18 @@ static int _unix_write(MainParam_t *mp, int needfilter, const char *unixFile)
 	if ((Target = SimpleFileOpen(0, 0, unixFile,
 				     O_WRONLY | O_CREAT | O_TRUNC,
 				     errmsg, 0, 0, 0))) {
-		ret = 0;
-		if(needfilter && arg->textmode){
-			Source = open_filter(COPY(File),arg->convertCharset);
-			if (!Source)
-				ret = -1;
-		} else
-			Source = COPY(File);
+		mt_off_t ret;
+		Source = COPY(File);
+		if(needfilter && arg->textmode)
+			Source = open_dos2unix(Source,arg->convertCharset);
 
-		if (ret == 0 )
+		if (Source)
 			ret = copyfile(Source, Target);
+		else
+			ret = -1;
 		FREE(&Source);
 		FREE(&Target);
-		if(ret <= -1){
+		if(ret < 0){
 			if(!arg->type)
 				unlink(unixFile);
 			return ERROR_ONE;
@@ -223,7 +308,7 @@ static int unix_copydir(direntry_t *entry, MainParam_t *mp)
 	if (!arg->recursive && mp->basenameHasWildcard)
 		return 0;
 
-	File->Class->get_data(File, &mtime, 0, 0, 0);	
+	File->Class->get_data(File, &mtime, 0, 0, 0);
 	if (!arg->preserveTime)
 		mtime = 0L;
 	if(!arg->type && arg->verbose) {
@@ -233,29 +318,28 @@ static int unix_copydir(direntry_t *entry, MainParam_t *mp)
 	}
 	if(got_signal)
 		return ERROR_ONE;
-	unixFile = mpBuildUnixFilename(mp);
+	unixFile = buildUnixFilename(arg);
 	if(!unixFile) {
 		printOom();
 		return ERROR_ONE;
 	}
-	if(arg->type || !*mpPickTargetName(mp) || !makeUnixDir(unixFile)) {
+	if(arg->type || !makeUnixDir(unixFile)) {
 		Arg_t newArg;
 
 		newArg = *arg;
 		newArg.mp.arg = (void *) &newArg;
-		newArg.mp.unixTarget = unixFile;
+		newArg.unixTarget = unixFile;
 		newArg.mp.targetName = 0;
 		newArg.mp.basenameHasWildcard = 1;
 
 		ret = mp->loop(File, &newArg.mp, "*");
 		set_mtime(unixFile, mtime);
 		free(unixFile);
-		return ret | GOT_ONE;		
+		return ret | GOT_ONE;
 	} else {
-		perror("mkdir");
-		fprintf(stderr, 
-			"Failure to make directory %s\n", 
-			unixFile);
+		fprintf(stderr,
+			"Failure to make directory %s: %s\n",
+			unixFile, strerror(errno));
 		free(unixFile);
 		return ERROR_ONE;
 	}
@@ -290,21 +374,20 @@ static int writeit(struct dos_name_t *dosname,
 	Stream_t *Target;
 	time_t now;
 	int type;
-	ssize_t ret;
+	mt_off_t ret;
 	uint32_t fat;
 	time_t date;
-	mt_size_t filesize, newsize;
+	mt_off_t filesize;
 	Arg_t *arg = (Arg_t *) arg0;
+	Stream_t *Source = COPY(arg->mp.File);
 
-
-
-	if (arg->mp.File->Class->get_data(arg->mp.File,
-									  & date, &filesize, &type, 0) < 0 ){
+	if (Source->Class->get_data(Source, &date, &filesize,
+				    &type, 0) < 0 ){
 		fprintf(stderr, "Can't stat source file\n");
 		return -1;
 	}
 
-	if(fileSizeTooBig(filesize)) {
+	if(fileTooBig(filesize)) {
 		fprintf(stderr, "File \"%s\" too big\n", longname);
 		return 1;
 	}
@@ -324,7 +407,7 @@ static int writeit(struct dos_name_t *dosname,
 	/* will it fit? */
 	if (!getfreeMinBytes(arg->mp.targetDir, filesize))
 		return -1;
-	
+
 	/* preserve mod time? */
 	if (arg->preserveTime)
 		now = date;
@@ -338,22 +421,19 @@ static int writeit(struct dos_name_t *dosname,
 		fprintf(stderr,"Could not open Target\n");
 		exit(1);
 	}
-	if (arg->needfilter & arg->textmode)
-		Target = open_filter(Target,arg->convertCharset);
-
-
-
-	ret = copyfile(arg->mp.File, Target);
-	GET_DATA(Target, 0, &newsize, 0, &fat);
+	if (arg->needfilter & arg->textmode) {
+		Source = open_unix2dos(Source,arg->convertCharset);
+	}
+		
+	ret = copyfile(Source, Target);
+	GET_DATA(Target, 0, 0, 0, &fat);
+	FREE(&Source);
 	FREE(&Target);
-	if (arg->needfilter & arg->textmode)
-	    newsize++; /* ugly hack: we gathered the size before the Ctrl-Z
-			* was written.  Increment it manually */
 	if(ret < 0 ){
 		fat_free(arg->mp.targetDir, fat);
 		return -1;
 	} else {
-		mk_entry(dosname, arg->attr, fat, (uint32_t)newsize,
+		mk_entry(dosname, arg->attr, fat, (uint32_t)ret,
 				 now, &entry->dir);
 		return 0;
 	}
@@ -462,9 +542,9 @@ static int dos_copydir(direntry_t *entry, MainParam_t *mp)
 		/* maybe the directory already exist. Use it */
 		newArg.mp.targetDir = subDir(mp->targetDir, targetName);
 		if(!newArg.mp.targetDir)
-			newArg.mp.targetDir = createDir(mp->targetDir, 
+			newArg.mp.targetDir = createDir(mp->targetDir,
 							targetName,
-							&arg->ch, arg->attr, 
+							&arg->ch, arg->attr,
 							now);
 	} else
 		newArg.mp.targetDir = mp->targetDir;
@@ -497,7 +577,7 @@ static void usage(int ret)
 	fprintf(stderr,
 		"Usage: %s [-spatnmQVBT] [-D clash_option] sourcefile targetfile\n", progname);
 	fprintf(stderr,
-		"       %s [-spatnmQVBT] [-D clash_option] sourcefile [sourcefiles...] targetdirectory\n", 
+		"       %s [-spatnmQVBT] [-D clash_option] sourcefile [sourcefiles...] targetdirectory\n",
 		progname);
 	exit(ret);
 }
@@ -507,7 +587,7 @@ void mcopy(int argc, char **argv, int mtype)
 {
 	Arg_t arg;
 	int c, fastquit;
-	
+
 
 	/* get command line options */
 
@@ -561,7 +641,7 @@ void mcopy(int argc, char **argv, int mtype)
 				batchmode = 1;
 				break;
 			case 'o':
-				handle_clash_options(&arg.ch, (char) c);
+				handle_clash_options(&arg.ch, c);
 				break;
 			case 'D':
 				if(handle_clash_options(&arg.ch, *optarg))
@@ -581,7 +661,9 @@ void mcopy(int argc, char **argv, int mtype)
 		usage(1);
 
 	init_mp(&arg.mp);
-	arg.mp.lookupflags = ACCEPT_PLAIN | ACCEPT_DIR | DO_OPEN | NO_DOTS;
+	arg.unixTarget = 0;
+	arg.mp.lookupflags =
+		ACCEPT_PLAIN | ACCEPT_DIR | DO_OPEN | NO_DOTS | DEFERABLE;
 	arg.mp.fast_quit = fastquit;
 	arg.mp.arg = (void *) &arg;
 	arg.mp.openflags = O_RDONLY;
@@ -596,10 +678,10 @@ void mcopy(int argc, char **argv, int mtype)
 	if(mtype){
 		/* Mtype = copying to stdout */
 		arg.mp.targetName = strdup("-");
-		arg.mp.unixTarget = strdup("");
+		arg.unixTarget = strdup("");
 		arg.mp.callback = dos_to_unix;
 		arg.mp.dirCallback = unix_copydir;
-		arg.mp.unixcallback = unix_to_unix;		
+		arg.mp.unixcallback = unix_to_unix;
 	} else {
 		const char *target;
 		if (argc - optind == 1) {
@@ -612,14 +694,18 @@ void mcopy(int argc, char **argv, int mtype)
 			target = argv[argc];
 		}
 
-		target_lookup(&arg.mp, target);
-		if(!arg.mp.targetDir && !arg.mp.unixTarget) {
+		if(target_lookup(&arg, target) == ERROR_ONE) {
+			fprintf(stderr,"%s: %s\n", target, strerror(errno));
+			exit(1);
+
+		}
+		if(!arg.mp.targetDir && !arg.unixTarget) {
 			fprintf(stderr,"Bad target %s\n", target);
 			exit(1);
 		}
 
 		/* callback functions */
-		if(arg.mp.unixTarget) {
+		if(arg.unixTarget) {
 			arg.mp.callback = dos_to_unix;
 			arg.mp.dirCallback = directory_dos_to_unix;
 			arg.mp.unixcallback = unix_to_unix;
