@@ -24,22 +24,20 @@
  */
 
 #include "sysincludes.h"
-#include "msdos.h"
 #include "mtools.h"
 #include "vfat.h"
 #include "nameclash.h"
 #include "fs.h"
 #include "stream.h"
-#include "mainloop.h"
 #include "file_name.h"
 
 /**
  * Converts input to shortname
  * @param un unix name (in Unix charset)
- * 
+ *
  * @return 1 if name had to be mangled
  */
-static __inline__ int convert_to_shortname(doscp_t *cp, ClashHandling_t *ch,
+static inline int convert_to_shortname(doscp_t *cp, ClashHandling_t *ch,
 					   const char *un, dos_name_t *dn)
 {
 	int mangled;
@@ -52,9 +50,9 @@ static __inline__ int convert_to_shortname(doscp_t *cp, ClashHandling_t *ch,
 	return mangled;
 }
 
-static __inline__ void chomp(char *line)
+static inline void chomp(char *line)
 {
-	int l = strlen(line);
+	size_t l = strlen(line);
 	while(l > 0 && (line[l-1] == '\n' || line[l-1] == '\r')) {
 		line[--l] = '\0';
 	}
@@ -63,14 +61,14 @@ static __inline__ void chomp(char *line)
 /**
  * Asks for an alternative new name for a file, in case of a clash
  */
-static __inline__ int ask_rename(doscp_t *cp, ClashHandling_t *ch,
+static inline int ask_rename(doscp_t *cp, ClashHandling_t *ch,
 				 dos_name_t *shortname,
 				 char *longname,
 				 int isprimary)
 {
 	int mangled;
 
-	/* TODO: Would be nice to suggest "autorenamed" version of name, press 
+	/* TODO: Would be nice to suggest "autorenamed" version of name, press
 	 * <Return> to get it.
 	 */
 #if 0
@@ -80,9 +78,6 @@ static __inline__ int ask_rename(doscp_t *cp, ClashHandling_t *ch,
 	if(!opentty(0))
 		return 0;
 
-#define maxsize (isprimary ?  MAX_VNAMELEN+1 : 11+1)
-#define name (isprimary ? argname : shortname)
-
 	mangled = 0;
 	do {
 		char tname[4*MAX_VNAMELEN+1];
@@ -95,12 +90,10 @@ static __inline__ int ask_rename(doscp_t *cp, ClashHandling_t *ch,
 		if (isprimary)
 			strcpy(longname, tname);
 		else
-			mangled = convert_to_shortname(cp, 
+			mangled = convert_to_shortname(cp,
 						       ch, tname, shortname);
 	} while (mangled & 1);
 	return 1;
-#undef maxsize
-#undef name
 }
 
 /**
@@ -109,10 +102,10 @@ static __inline__ int ask_rename(doscp_t *cp, ClashHandling_t *ch,
  * The decision either comes from the default (ch), or the user will be
  * prompted if there is no default
  */
-static __inline__ clash_action ask_namematch(doscp_t *cp,
+static inline clash_action ask_namematch(doscp_t *cp,
 					     dos_name_t *dosname,
 					     char *longname,
-					     int isprimary, 
+					     int isprimary,
 					     ClashHandling_t *ch,
 					     int no_overwrite,
 					     int reason)
@@ -161,7 +154,8 @@ static __inline__ clash_action ask_namematch(doscp_t *cp,
 	}
 
 	if (!isprimary)
-		name = unix_normalize(cp, name_buffer, dosname);
+		name = unix_normalize(cp, name_buffer,
+				      dosname, sizeof(*dosname));
 	else
 		name = longname;
 
@@ -182,12 +176,12 @@ static __inline__ clash_action ask_namematch(doscp_t *cp,
 		fflush(opentty(1));
 		if (mtools_raw_tty) {
 			int rep;
-			rep = fgetc(opentty(1));			
+			rep = fgetc(opentty(1));
 			fputs("\n", stderr);
 			if(rep == EOF)
 				ans[0] = 'q';
 			else
-				ans[0] = rep;
+				ans[0] = (char) rep;
 		} else {
 			if(fgets(ans, 9, opentty(0)) == NULL)
 				ans[0] = 'q';
@@ -251,7 +245,7 @@ static __inline__ clash_action ask_namematch(doscp_t *cp,
  *
  * Also, immediately copy the original name so that messages can use it.
  */
-static __inline__ clash_action process_namematch(doscp_t *cp,
+static inline clash_action process_namematch(doscp_t *cp,
 						 dos_name_t *dosname,
 						 char *longname,
 						 int isprimary,
@@ -303,9 +297,13 @@ static __inline__ clash_action process_namematch(doscp_t *cp,
 			return NAMEMATCH_SKIP;
 		else
 			return NAMEMATCH_OVERWRITE;
-	default:
+	case NAMEMATCH_NONE:
+	case NAMEMATCH_ERROR:
+	case NAMEMATCH_SUCCESS:
+	case NAMEMATCH_GREW:
 		return NAMEMATCH_NONE;
 	}
+	return action;
 }
 
 static int contains_illegals(const char *string, const char *illegals,
@@ -336,11 +334,11 @@ static int is_reserved(char *ans, int islong)
 		    ((islong && !ans[4]) ||
 		     (!islong && !strncmp(ans+4,"    ",4))))
 			return 1;
-	
+
 	return 0;
 }
 
-static __inline__ clash_action get_slots(Stream_t *Dir,
+static inline clash_action get_slots(Stream_t *Dir,
 					 dos_name_t *dosname,
 					 char *longname,
 					 struct scan_state *ssp,
@@ -388,26 +386,25 @@ static __inline__ clash_action get_slots(Stream_t *Dir,
 					ch->use_longname)) {
 			case -1:
 				return NAMEMATCH_ERROR;
-				
+
 			case 0:
 				return NAMEMATCH_SKIP;
 				/* Single-file error error or skip request */
-				
+
 			case 5:
 				return NAMEMATCH_GREW;
 				/* Grew directory, try again */
-				
+
 			case 6:
 				return NAMEMATCH_SUCCESS; /* Success */
-		}	
-		match_pos = -2;
-		if (ssp->longmatch > -1) {
+		}
+		if (ssp->longmatch >= 0) {
 			/* Primary Long Name Match */
 #ifdef debug
 			fprintf(stderr,
 				"Got longmatch=%d for name %s.\n",
 				longmatch, longname);
-#endif			
+#endif
 			match_pos = ssp->longmatch;
 			isprimary = 1;
 		} else if ((ch->use_longname & 1) && (ssp->shortmatch != -1)) {
@@ -419,6 +416,15 @@ static __inline__ clash_action get_slots(Stream_t *Dir,
 #endif
 
 			match_pos = ssp->shortmatch;
+			/* match_pos may become negative here (-2) in case
+			 * of pessimisticShortRename, i.e. creating a
+			 * long entry whose shortname matches another
+			 * entry's shortname:
+			 * mformat -C b: -s 18 -h 2 -t 80
+			 * mcopy /etc/issue b:12345678a
+			 * mcopy /etc/issue b:12345678b
+			 */
+			
 			isprimary = 0;
 		} else if (ssp->shortmatch >= 0) {
 			/* Primary Short Name Match */
@@ -443,7 +449,7 @@ static __inline__ clash_action get_slots(Stream_t *Dir,
 	}
 	ret = process_namematch(cp, dosname, longname,
 				isprimary, ch, no_overwrite, reason);
-	
+
 	if (ret == NAMEMATCH_OVERWRITE && match_pos > -1){
 		if((entry.dir.attr & 0x5) &&
 		   (ask_confirmation("file is read only, overwrite anyway (y/n) ? ")))
@@ -451,7 +457,7 @@ static __inline__ clash_action get_slots(Stream_t *Dir,
 		/* Free up the file to be overwritten */
 		if(fatFreeWithDirentry(&entry))
 			return NAMEMATCH_ERROR;
-		
+
 #if 0
 		if(isprimary &&
 		   match_pos - ssp->match_free + 1 >= ssp->size_needed){
@@ -475,7 +481,7 @@ static __inline__ clash_action get_slots(Stream_t *Dir,
 }
 
 
-static __inline__ int write_slots(Stream_t *Dir,
+static inline int write_slots(Stream_t *Dir,
 				  dos_name_t *dosname,
 				  char *longname,
 				  struct scan_state *ssp,
@@ -490,7 +496,8 @@ static __inline__ int write_slots(Stream_t *Dir,
 		return 0;
 
 	entry.Dir = Dir;
-	entry.entry = ssp->slot;
+	assert(ssp->got_slots);
+	setEntryToPos(&entry, ssp->slot);
 	native_to_wchar(longname, entry.name, MAX_VNAMELEN, 0, 0);
 	entry.name[MAX_VNAMELEN]='\0';
 	entry.dir.Case = Case & (EXTCASE | BASECASE);
@@ -504,8 +511,6 @@ static __inline__ int write_slots(Stream_t *Dir,
 			write_vfat(Dir, dosname, 0,
 				   ssp->free_start, &entry);
 		}
-		/* clear_vses(Dir, ssp->free_start + ssp->size_needed,
-		   ssp->free_end); */
 	} else
 		return 0;
 
@@ -591,7 +596,7 @@ static int _mwrite_one(Stream_t *Dir,
 			case NAMEMATCH_ERROR:
 				return -1;	/* Non-file-specific error,
 						 * quit */
-				
+
 			case NAMEMATCH_SKIP:
 				return -1;	/* Skip file (user request or
 						 * error) */
@@ -617,7 +622,7 @@ static int _mwrite_one(Stream_t *Dir,
 					return -1;
 				}
 				expanded = 1;
-				
+
 				if (dir_grow(Dir, scan.max_entry))
 					return -1;
 				continue;
@@ -626,7 +631,9 @@ static int _mwrite_one(Stream_t *Dir,
 				return write_slots(Dir, &dosname, longname,
 						   &scan, cb, arg,
 						   ch->use_longname);
-			default:
+			case NAMEMATCH_NONE:
+			case NAMEMATCH_AUTORENAME:
+			case NAMEMATCH_QUIT:
 				fprintf(stderr,
 					"Internal error: clash_action=%d\n",
 					ret);
@@ -675,7 +682,7 @@ void init_clash_handling(ClashHandling_t *ch)
 	ch->is_label = 0;
 }
 
-int handle_clash_options(ClashHandling_t *ch, char c)
+int handle_clash_options(ClashHandling_t *ch, int c)
 {
 	int isprimary;
 	if(isupper(c))
